#include "ChipSize.h"

#include <math.h>
#include <mathconst/mathconst.h>

ChipSize& ChipSize::operator = ( const ChipSize& rhs ) {

  // check for assignment against itself
  if ( this == &rhs )
    return *this;

  name = rhs.name;
  id   = rhs.id;

  for ( size_t ii = 0; ii < 2; ii++ ) {

    chip_size_pix[ ii ] = rhs.chip_size_pix[ ii ];
    pixel_size_mm[ ii ] = rhs.pixel_size_mm[ ii ];
    chip_size_mm[ ii ]  = rhs.chip_size_mm[ ii ];

  }

  theta = rhs.theta;
  delta = rhs.delta;
  X0    = rhs.X0;
  Y0    = rhs.Y0;
  H     = rhs.H;
  cos_theta = rhs.cos_theta;
  sin_theta = rhs.sin_theta;

  return *this;
}

ChipSize::ChipSize( const char* header[], RDBColumn** rdb_column )
  throw( Exception ) {

  try {

    init_ChipSize( header, rdb_column );

  } catch( Exception& e ) {

    throw;

  }

}

void ChipSize::init_ChipSize( const char* header[],
			      RDBColumn** rdb_column )
  throw( Exception ) {

  try {

    for( size_t num = 0; NULL != header[ num ]; num++ ) {

      if ( 0 == strcmp( header[ num ], "chip" ) )
	simpleRDBTable< ChipSize >::getData( rdb_column[ num ], name );
      else if ( 0 == strcmp( header[ num ], "id" ) )
	simpleRDBTable< ChipSize >::getData( rdb_column[ num ], id );
      else if ( 0 == strcmp( header[ num ], "xsize_pix" ) )
	simpleRDBTable< ChipSize >::getData( rdb_column[ num ],
					     chip_size_pix[ XXX ] );
      else if ( 0 == strcmp( header[ num ], "ysize_pix" ) )
	simpleRDBTable< ChipSize >::getData( rdb_column[ num ],
					     chip_size_pix[ YYY ] );
      else if ( 0 == strcmp( header[ num ], "xpixelsize_micron" ) ) {
	simpleRDBTable< ChipSize >::getData( rdb_column[ num ],
					     pixel_size_mm[ XXX ] );
	pixel_size_mm[ XXX ] /= 1000.0;
      } else if ( 0 == strcmp( header[ num ], "ypixelsize_micron" ) ) {
	simpleRDBTable< ChipSize >::getData( rdb_column[ num ],
					     pixel_size_mm[ YYY ] );
	pixel_size_mm[ YYY ] /= 1000.0;
      } else if ( 0 == strcmp( header[ num ], "xsize_mm" ) )
	simpleRDBTable< ChipSize >::getData( rdb_column[ num ],
					     chip_size_mm[ XXX ] );
      else if ( 0 == strcmp( header[ num ], "ysize_mm" ) )
	simpleRDBTable< ChipSize >::getData( rdb_column[ num ],
					     chip_size_mm[ YYY ] );
      else if ( 0 == strcmp( header[ num ], "theta" ) ) {
	simpleRDBTable< ChipSize >::getData( rdb_column[ num ], theta );
	cos_theta = cos( theta * M_PI / 180 );
	sin_theta = sin( theta * M_PI / 180 );
      } else if ( 0 == strcmp( header[ num ], "delta" ) )
	simpleRDBTable< ChipSize >::getData( rdb_column[ num ], delta );
      else if ( 0 == strcmp( header[ num ], "X0" ) )
	simpleRDBTable< ChipSize >::getData( rdb_column[ num ], X0 );
      else if ( 0 == strcmp( header[ num ], "Y0" ) )
	simpleRDBTable< ChipSize >::getData( rdb_column[ num ], Y0 );
      else if ( 0 == strcmp( header[ num ], "H" ) )
	simpleRDBTable< ChipSize >::getData( rdb_column[ num ], H );

      else {

	throw Exception( "ChipSize::init_ChipSize( const char**, RDBColumn**) "
			 ": Unrecognized option\n" );

      }

    }

  } catch( Exception& e ) {

    throw;

  }

}

void ChipSize::print( ostream& os ) const {

  os << name << '\t';
  os << chip_size_pix[ 0 ] << '\t';
  os << chip_size_pix[ 1 ] << '\t';
  os << pixel_size_mm[ 0 ] << '\t';
  os << pixel_size_mm[ 1 ] << '\t';
  os << chip_size_mm[ 0 ] << '\t';
  os << chip_size_mm[ 1 ];

}

/*
dumbo-70: maple
    |\^/|     Maple V Release 2 (Harvard University)
._|\|   |/|_. Copyright (c) 1981-1992 by the University of Waterloo.
 \  MAPLE  /  All rights reserved. MAPLE is a registered trademark of
 <____ ____>  Waterloo Maple Software.
      |       Type ? for help.
> A:=array([ [1,0], [0,H] ] );

                                      [ 1  0 ]
                                 A := [      ]
                                      [ 0  H ]

> B:=array([ [cos(theta),sin(theta)], [-sin(theta),cos(theta)] ] );

                            [  cos(theta)   sin(theta) ]
                       B := [                          ]
                            [ - sin(theta)  cos(theta) ]

> C:=array([ [ chipx_half ], [ chipy_half ] ] );

                                   [ chipx_half ]
                              C := [            ]
                                   [ chipy_half ]

> evalm( A &* B &* C );

            [    cos(theta) chipx_half + sin(theta) chipy_half    ]
            [                                                     ]
            [ - H sin(theta) chipx_half + H cos(theta) chipy_half ]

 */
double ChipSize::tdeltx( const double chipx, const double chipy ) const {

  double chipx_half = chipx - 0.5;
  double chipy_half = chipy - 0.5;
  return delta * ( cos_theta * chipx_half + sin_theta * chipy_half ) + X0 + 0.5;

}

double ChipSize::tdelty( const double chipx, const double chipy ) const {

  double chipx_half = chipx - 0.5;
  double chipy_half = chipy - 0.5;
  return delta * ( -H * sin_theta * chipx_half + H * cos_theta * chipy_half ) + Y0 + 0.5;

}
