#ifndef ChipSize_H
#define ChipSize_H

#include <iostream>
#include <map>

#include <Exception/Exception.h>

#include <rdbxx/simpleRDBTable.h>

using namespace std;

class ChipSize {

  friend ostream& operator << ( ostream& os, ChipSize& a ) {
    a.print( os );
    return os;
  }

  friend ostream& operator << ( ostream& os, ChipSize* a ) {
    a->print( os );
    return os;
  }

public:

  enum Pos { XXX, YYY };

  virtual ~ChipSize( ) { }

  ChipSize( ) { }

  ChipSize( const char* header[], RDBColumn** rdb_column )
    throw( Exception );

  ChipSize( const ChipSize& c ) { *this = c; }

  ChipSize& operator = ( const ChipSize& rhs );

  void init_ChipSize( const char* header[], RDBColumn** rdb_column )
    throw( Exception );

  string get_name( void ) const { return name; }

  int    get_chip_id( ) const { return id; }

  int get_chip_size_pix( Pos pos ) const { return chip_size_pix[ pos ]; }

  double get_pixel_size_mm( Pos pos ) const {
    return pixel_size_mm[ pos ]; }

  double get_chip_size_mm( Pos pos ) const { return chip_size_mm[ pos ]; }

  double tdeltx( const double chipx, const double chipy ) const;

  double tdelty( const double chipx, const double chipy ) const;
  
  void print( ostream& os ) const;

protected:

  string name;
  int id;

  int chip_size_pix[ 2 ];

  // The database is given in micron, will store it as mm.
  double pixel_size_mm[ 2 ];

  double chip_size_mm[ 2 ];

  double theta;
  double delta;
  double X0;
  double Y0;
  double H;
  double cos_theta;
  double sin_theta;

};

typedef map< string, ChipSize, less< string > > ChipSizeMap;
typedef ChipSizeMap::const_iterator ChipSizeMapIterator;

#endif
