/*

objs/sun4u-SunOS-5/deticpt chipdb=a.rdb input=input output=stdout | bp2rdb input=stdin output=stdout | rdbplt input=stdin xcol=cpcpos_x ycol=cpcpos_y connect- just+ marker=-1

 */
#include <stdlib.h>
#include <libgen.h>

#include <suplibxx/str.h>
#include <paramxx/ParFile.h>
#include <dvm3/dvm3_vector.h>
#include <rl_basicray/rl_BasicRay.h>
#include <tracefctxx/TraceFct.h>
#include <BinaryPipe/BinaryPipe.h>
#include <BinaryPipe/MyPhoton.h>
#include <rdbxx/RDBColumn.h>
#include <rdbxx/simpleRDBTable.h>

#include "ChipSize.h"
#include "Detector.h"
#include "config.h"

/**
 * The data packet to add
 */
DataPacket datapacket_pos_is_pixelized[] = {
    { "position_np", BPDType_DVector3, NULL },
    { "cpcpos",      BPDType_DVector3, NULL },
    { "cpcpos_np",   BPDType_DVector3, NULL },
    { "cpcdir",      BPDType_DVector3, NULL },
    { "chip",        BPDType_IVector2, NULL },
    { "chip_np",     BPDType_DVector2, NULL },
    { "tdet",        BPDType_IVector2, NULL },
    { "tdet_np",     BPDType_DVector2, NULL },
    { "det",         BPDType_DVector2, NULL },
    { "det_np",      BPDType_DVector2, NULL },
    { "chip_id",     BPDType_int,      NULL },
    { "icpt",        BPDType_int,      NULL },
    { NULL,          BPDType_char,     NULL }
};
DataPacket datapacket_pos_not_pixelized[] = {
    { "position_pix", BPDType_DVector3, NULL },
    { "cpcpos_pix",   BPDType_DVector3, NULL },
    { "cpcpos",       BPDType_DVector3, NULL },
    { "cpcdir",       BPDType_DVector3, NULL },
    { "chip_pix",     BPDType_IVector2, NULL },
    { "chip",         BPDType_DVector2, NULL },
    { "tdet_pix",     BPDType_IVector2, NULL },
    { "tdet",         BPDType_DVector2, NULL },
    { "det_pix",      BPDType_DVector2, NULL },
    { "det",          BPDType_DVector2, NULL },
    { "chip_id",      BPDType_int,      NULL },
    { "icpt",         BPDType_int,      NULL },
    { NULL,           BPDType_char,     NULL }
};
DataPacket* dp;

/**
 *
 */
class DeticptPhoton : public MyPhoton {

public:

  DVector3* get_position( )    { return position; }
  DVector3* get_position_np( ) { return position_np; }
  DVector3* get_cpcpos( )      { return cpcpos; }
  DVector3* get_cpcpos_np( )   { return cpcpos_np; }
  DVector3* get_cpcdir( )      { return cpcdir; }
  IVector2* get_chip( )        { return chip; }
  DVector2* get_chip_np( )     { return chip_np; }
  IVector2* get_tdet( )        { return tdet; }
  DVector2* get_tdet_np( )     { return tdet_np; }
  DVector2* get_det( )         { return  det; }
  DVector2* get_det_np( )      { return  det_np; }
  int*      get_chip_id( )       { return chip_id; }
  int*      get_icpt( )          { return icpt; }
  DeticptPhoton( BPipe* bpipe, void* core ) throw ( BinaryPipeException ) :
    MyPhoton( bpipe, core ) {

    position    = (DVector3*) dpktf_data( bpipe, core, (char*) "position" );
    position_np	= (DVector3*) dpktf_data( bpipe, core, (char*) dp[0].name );
    cpcpos	= (DVector3*) dpktf_data( bpipe, core, (char*) dp[1].name );
    cpcpos_np	= (DVector3*) dpktf_data( bpipe, core, (char*) dp[2].name );
    cpcdir	= (DVector3*) dpktf_data( bpipe, core, (char*) dp[3].name );
    chip	= (IVector2*) dpktf_data( bpipe, core, (char*) dp[4].name );
    chip_np	= (DVector2*) dpktf_data( bpipe, core, (char*) dp[5].name );
    tdet	= (IVector2*) dpktf_data( bpipe, core, (char*) dp[6].name );
    tdet_np	= (DVector2*) dpktf_data( bpipe, core, (char*) dp[7].name );
    det		= (DVector2*) dpktf_data( bpipe, core, (char*) dp[8].name );
    det_np	= (DVector2*) dpktf_data( bpipe, core, (char*) dp[9].name );
    chip_id	= (int*) dpktf_data( bpipe, core, (char*) dp[10].name );
    icpt        = (int*) dpktf_data( bpipe, core, (char*) dp[11].name );
  }
  void set_result( DeticptResult& deticptResult ) {
    deticptResult.position    = *(position);
    deticptResult.position_np = *(position_np);
    deticptResult.cpcpos      = *(cpcpos);
    deticptResult.cpcpos_np   = *(cpcpos_np);
    deticptResult.cpcdir      = *(cpcdir);
    deticptResult.chip	      = *(chip);
    deticptResult.chip_np     = *(chip_np);
    deticptResult.tdet	      = *(tdet);
    deticptResult.tdet_np     = *(tdet_np);
    deticptResult.det	      = *(det);
    deticptResult.det_np      = *(det_np);
    deticptResult.chip_id     = *(chip_id);
    deticptResult.icpt        = *(icpt);
  }
  void set_photon( DeticptResult& deticptResult, const bool pixprim ) {
    *(position)    = (pixprim?deticptResult.position:deticptResult.position_np);
    *(position_np) = (pixprim?deticptResult.position_np:deticptResult.position);
    *(cpcpos)      = deticptResult.cpcpos;
    *(cpcpos_np)   = deticptResult.cpcpos_np;
    *(cpcdir)      = deticptResult.cpcdir;
    *(chip)        = deticptResult.chip;
    *(chip_np)     = deticptResult.chip_np;
    *(tdet)        = deticptResult.tdet;
    *(tdet_np)     = deticptResult.tdet_np;
    *(det)         = deticptResult.det;
    *(det_np)      = deticptResult.det_np;
    *(chip_id)     = deticptResult.chip_id;
    *(icpt)        = deticptResult.icpt;
  }

protected:
  DVector3* position;
  DVector3* position_np;
  DVector3* cpcpos;
  DVector3* cpcpos_np;
  DVector3* cpcdir;
  IVector2* chip;
  DVector2* chip_np;
  IVector2* tdet;
  DVector2* tdet_np;
  DVector2* det;
  DVector2* det_np;
  int* chip_id;
  int* icpt;

};

int usage( );

void one_of_three_ways( string parvalue, double& value ) {
  if ( parvalue != "default" ) {
    if ( 'f' == parvalue[ 0 ] ) { // Offset relative to aimpointdb
      parvalue.erase( 0, parvalue.find_first_not_of( "f " ) );
      value = value + suplib::str2d( parvalue.c_str( ) );
    } 
    else { // Absolute position, override aimpointdb
      value = suplib::str2d( parvalue.c_str( ) );
    }
  }
}

// The following procedure function should be made into a container class.
void readChipSizeMapDataBase( const string& fname, const char* header[],
			      ChipSizeMap& csm ) throw ( Exception ) {

  try {

    simpleRDBTable< ChipSize > my_rdb( fname, ios::in, header );
    ChipSize* chip_size = NULL;
    while( NULL != (chip_size=my_rdb.readRow( ) ) ) {

      string key( chip_size->get_name( ) );

      ChipSizeMapIterator chipMapIterator = csm.find( key );
      if ( chipMapIterator == csm.end( ) ) {

	csm[ key ] = *chip_size;

      } else {

	cerr << "Duplicate entry for key `" << key << "' in file `";
	cerr << fname << "'\n";
	delete chip_size;
	exit( EXIT_FAILURE );

      }

      delete chip_size;
    }

    /*
    cerr << my_rdb << endl;
    for ( ChipSizeMapIterator iter = csm.begin( ); iter != csm.end( );
	  ++iter ) {
      iter->second.print( cerr );
      cerr << endl;
    }
    */

  } catch( RDBErr& re ) {
    re.set_message( string("error reading '") + fname + "'" );
    throw( re );

  } 
}

void readAimPointDataBase( const string& fname, dvm3_Vector& aimpointOSAC, dvm3_Vector& nodeOSAC, double& focal_length ) throw ( Exception ) {

  try {

    RDB rdbtable( fname, ios::in );

    aimpointOSAC[0] = suplib::str2d( rdbtable.getComment( "xaimpoint" ).getValue().c_str( ) );
    aimpointOSAC[1] = suplib::str2d( rdbtable.getComment( "yaimpoint" ).getValue().c_str( ) );
    aimpointOSAC[2] = suplib::str2d( rdbtable.getComment( "zaimpoint" ).getValue().c_str( ) );
    nodeOSAC[0]     = suplib::str2d( rdbtable.getComment( "xnode" ).getValue().c_str( ) );
    nodeOSAC[1]     = suplib::str2d( rdbtable.getComment( "ynode" ).getValue().c_str( ) );
    nodeOSAC[2]     = suplib::str2d( rdbtable.getComment( "znode" ).getValue().c_str( ) );
    focal_length    = suplib::str2d( rdbtable.getComment( "flength" ).getValue().c_str( ) );


  } catch( RDBErr& re ) {
    re.set_message( ("error reading '") + fname + "'" );
    throw( re );

  } 
}

void deticpt( ParFile& parfile )
  throw( ParFileException, DeticptException, BinaryPipeException, Exception ) {

  try {

    const char* rdbheader[] = { "xcpc", "ycpc", "zcpc",
				  "xcpcclip", "ycpcclip",
				  "xlsi", "ylsi", "zlsi", 
				  "chip", "corner", NULL };

    const char* corners_id[] = { "LL", "LR", "UR", "UL", NULL };

    const char* rdb_size_header[] = { "chip", "id",
                                      "xsize_pix",  "ysize_pix",
                                      "xpixelsize_micron", "ypixelsize_micron",
                                      "xsize_mm",   "ysize_mm",
				      "theta", "delta", "X0", "Y0", "H",
				      NULL };

    ChipSizeMap chipSizeMap;
    readChipSizeMapDataBase( parfile.pgetstring( "chipsizedb" ),
			     rdb_size_header, chipSizeMap );

    Detector detector( parfile, rdbheader, corners_id, chipSizeMap );
    //    cerr << detector << endl;
    if ( parfile.pgetb( "diagnostic" ) ) {
      detector.diagnostics_only( cout );
      exit( EXIT_SUCCESS );
    }

    bool pixprim = parfile.pgetb( "pixprim" );
    if ( pixprim ) 
      dp = datapacket_pos_is_pixelized;
    else 
      dp = datapacket_pos_not_pixelized;

    BinaryPipe< DeticptPhoton > bpipe( parfile.pgetstring( "input" ).c_str( ),
				       parfile.pgetstring( "output" ).c_str( ),
				       NULL, dp );

    DeticptPhoton* photon     = (DeticptPhoton*) bpipe;
    DVector3*  direction      = photon->get_direction( );
    DeticptResult deticptResult;

    dvm3_Vector aimpointOSAC, nodeOSAC;
    double focal_length;

    readAimPointDataBase( parfile.pgetstring( "aimpointdb" ), aimpointOSAC, nodeOSAC, focal_length );

    one_of_three_ways( parfile.pgetstring( "xaimpoint" ),  aimpointOSAC[0] );
    one_of_three_ways( parfile.pgetstring( "yaimpoint" ),  aimpointOSAC[1] );
    one_of_three_ways( parfile.pgetstring( "zaimpoint" ),  aimpointOSAC[2] );

    // aimoffset is done in Detector since it has access to the
    // chipposdb default values...

    one_of_three_ways( parfile.pgetstring( "xnode" ),      nodeOSAC[0]     );
    one_of_three_ways( parfile.pgetstring( "ynode" ),      nodeOSAC[1]     );
    one_of_three_ways( parfile.pgetstring( "znode" ),      nodeOSAC[2]     );

    one_of_three_ways( parfile.pgetstring( "flength" ),    focal_length    );

    bool filter = parfile.pgetb( "filter" );

    MirrorPars mirrorPars( focal_length );

    while ( bpipe.read_dpkts( ) ) {

      photon->set_result( deticptResult );

      /* Direction and position of photon in OSAC coords */
      deticptResult.icpt = 0;

      dvm3_Vector dir( direction->x, direction->y, direction->z );
      dvm3_Vector pos( deticptResult.position.x, deticptResult.position.y, deticptResult.position.z);

      if ( EXIT_SUCCESS == detector.project_to_detector( dir, 
							 pos,
							 mirrorPars, 
							 aimpointOSAC,
							 nodeOSAC,
							 deticptResult
							 ) ) 
	deticptResult.icpt = 1;

      if ( deticptResult.icpt || filter ) {
	photon->set_photon( deticptResult, pixprim );
	if ( bpipe.write_dpkt( ) )
	  throw( DeticptException( bpipe_strerror( bpipe_errno ) ) );
      }

      // cout << *photon << endl;

    }

  } catch( RDBErr& re ) {
    throw( re );

  } catch( BinaryPipeException& bpe ) {
    throw( bpe );

  } catch ( ParFileException& pfe ) {
    throw( pfe );

  } catch ( DeticptException& de ) {
    throw( de );

  } catch ( Exception& e ) {
    throw( e );

  } catch ( ... ) {
    throw( Exception( "Unknown exception caught\n" ) );

  }

}

int main( int argc, char** argv ) {
  TraceFct tf( argv[0], 0, -1 );

  try {

    ParFile parfile( argc, argv );
    // cout << parfile << endl;

    if ( parfile.pgetb( "help" ) ) {
      usage( );
      return EXIT_SUCCESS;
    }
    
    if ( parfile.pgetb( "version" ) ) {
      cout << basename( argv[ 0 ] ) << '\t' << VERSION << '\n';
      return EXIT_SUCCESS;
    }

    deticpt( parfile );

  } catch(  ParFileException& pfe ) {
    tf.exit( EXIT_FAILURE, pfe );

  } catch ( DeticptException& de ) {
    tf.exit( EXIT_FAILURE, de );

  } catch ( Exception& e ) {
    tf.exit( EXIT_FAILURE, e );

  }

  return EXIT_SUCCESS;

}
